"use strict";
var EntityMetadata_1 = require("../metadata/EntityMetadata");
var ColumnMetadata_1 = require("../metadata/ColumnMetadata");
var ForeignKeyMetadata_1 = require("../metadata/ForeignKeyMetadata");
var JunctionTableMetadata_1 = require("../metadata/JunctionTableMetadata");
var typeorm_1 = require("../typeorm");
var EntityMetadataValidator_1 = require("./EntityMetadataValidator");
/**
 * Aggregates all metadata: table, column, relation into one collection grouped by tables for a given set of classes.
 *
 * @internal
 */
var EntityMetadataBuilder = (function () {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    function EntityMetadataBuilder(namingStrategy) {
        this.namingStrategy = namingStrategy;
        // todo: type in function validation, inverse side function validation
        // todo: check on build for duplicate names, since naming checking was removed from MetadataStorage
        // todo: duplicate name checking for: table, relation, column, index, naming strategy, join tables/columns?
        this.metadataStorage = typeorm_1.defaultMetadataStorage();
        this.entityValidator = new EntityMetadataValidator_1.EntityMetadataValidator();
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Builds a complete metadata aggregations for the given entity classes.
     */
    EntityMetadataBuilder.prototype.build = function (entityClasses) {
        var _this = this;
        var allMetadataStorage = typeorm_1.defaultMetadataStorage();
        // filter the metadata only we need - those which are bind to the given table classes
        var allTableMetadatas = allMetadataStorage.tableMetadatas.filterByClasses(entityClasses);
        var tableMetadatas = allTableMetadatas.filterByClasses(entityClasses).filter(function (table) { return !table.isAbstract; });
        var entityMetadatas = tableMetadatas.map(function (tableMetadata) {
            var mergedMetadata = allMetadataStorage.mergeWithAbstract(allTableMetadatas, tableMetadata);
            var entityMetadata = new EntityMetadata_1.EntityMetadata(tableMetadata, mergedMetadata.columnMetadatas, mergedMetadata.relationMetadatas, mergedMetadata.indexMetadatas, mergedMetadata.compoundIndexMetadatas, []);
            // find entity's relations join tables
            entityMetadata.relations.forEach(function (relation) {
                var relationJoinTable = mergedMetadata.joinTableMetadatas.find(function (joinTable) { return joinTable.propertyName === relation.propertyName; });
                if (relationJoinTable)
                    relation.joinTable = relationJoinTable;
            });
            // find entity's relations join columns
            entityMetadata.relations.forEach(function (relation) {
                var relationJoinColumn = mergedMetadata.joinColumnMetadatas.find(function (joinColumn) { return joinColumn.propertyName === relation.propertyName; });
                if (relationJoinColumn)
                    relation.joinColumn = relationJoinColumn;
            });
            // set naming strategies
            tableMetadata.namingStrategy = _this.namingStrategy;
            entityMetadata.columns.forEach(function (column) { return column.namingStrategy = _this.namingStrategy; });
            entityMetadata.relations.forEach(function (relation) { return relation.namingStrategy = _this.namingStrategy; });
            return entityMetadata;
        });
        // set inverse side (related) entity metadatas for all relation metadatas
        entityMetadatas.forEach(function (entityMetadata) {
            entityMetadata.relations.forEach(function (relation) {
                relation.relatedEntityMetadata = entityMetadatas.find(function (m) { return m.target === relation.type; });
            });
        });
        // check for errors in build metadata schema (we need to check after relationEntityMetadata is set)
        this.entityValidator.validateMany(entityMetadatas);
        // generate columns and foreign keys for tables with relations
        entityMetadatas.forEach(function (metadata) {
            var foreignKeyRelations = metadata.ownerOneToOneRelations.concat(metadata.manyToOneRelations);
            foreignKeyRelations.map(function (relation) {
                var inverseSideMetadata = entityMetadatas.find(function (metadata) { return metadata.target === relation.type; });
                // find relational columns and if it does not exist - add it
                var relationalColumn = metadata.columns.find(function (column) { return column.name === relation.name; });
                if (!relationalColumn) {
                    var options = {
                        type: inverseSideMetadata.primaryColumn.type,
                        oldColumnName: relation.oldColumnName,
                        nullable: relation.isNullable
                    };
                    relationalColumn = new ColumnMetadata_1.ColumnMetadata({
                        target: metadata.target,
                        propertyName: relation.name,
                        propertyType: inverseSideMetadata.primaryColumn.type,
                        isVirtual: true,
                        options: options
                    });
                    metadata.columns.push(relationalColumn);
                }
                // create and add foreign key
                var foreignKey = new ForeignKeyMetadata_1.ForeignKeyMetadata(metadata.table, [relationalColumn], inverseSideMetadata.table, [inverseSideMetadata.primaryColumn], relation.onDelete);
                metadata.foreignKeys.push(foreignKey);
            });
        });
        // generate junction tables with its columns and foreign keys
        var junctionEntityMetadatas = [];
        entityMetadatas.forEach(function (metadata) {
            metadata.ownerManyToManyRelations.map(function (relation) {
                var inverseSideMetadata = entityMetadatas.find(function (metadata) { return metadata.target === relation.type; });
                var tableName = metadata.table.name + "_" + relation.name + "_" +
                    inverseSideMetadata.table.name + "_" + inverseSideMetadata.primaryColumn.name;
                var tableMetadata = new JunctionTableMetadata_1.JunctionTableMetadata(tableName);
                var column1options = {
                    length: metadata.primaryColumn.length,
                    type: metadata.primaryColumn.type,
                    name: metadata.table.name + "_" + metadata.primaryColumn.name + "_1"
                };
                var column2options = {
                    length: inverseSideMetadata.primaryColumn.length,
                    type: inverseSideMetadata.primaryColumn.type,
                    name: inverseSideMetadata.table.name + "_" + inverseSideMetadata.primaryColumn.name + "_2"
                };
                var columns = [
                    new ColumnMetadata_1.ColumnMetadata({
                        propertyType: inverseSideMetadata.primaryColumn.type,
                        options: column1options
                    }),
                    new ColumnMetadata_1.ColumnMetadata({
                        propertyType: inverseSideMetadata.primaryColumn.type,
                        options: column2options
                    })
                ];
                var foreignKeys = [
                    new ForeignKeyMetadata_1.ForeignKeyMetadata(tableMetadata, [columns[0]], metadata.table, [metadata.primaryColumn]),
                    new ForeignKeyMetadata_1.ForeignKeyMetadata(tableMetadata, [columns[1]], inverseSideMetadata.table, [inverseSideMetadata.primaryColumn]),
                ];
                var junctionEntityMetadata = new EntityMetadata_1.EntityMetadata(tableMetadata, columns, [], [], [], foreignKeys);
                junctionEntityMetadatas.push(junctionEntityMetadata);
                relation.junctionEntityMetadata = junctionEntityMetadata;
                if (relation.inverseRelation)
                    relation.inverseRelation.junctionEntityMetadata = junctionEntityMetadata;
            });
        });
        return entityMetadatas.concat(junctionEntityMetadatas);
    };
    return EntityMetadataBuilder;
}());
exports.EntityMetadataBuilder = EntityMetadataBuilder;
//# sourceMappingURL=EntityMetadataBuilder.js.map