"use strict";
var Rx = require("rxjs/Rx");
/**
 * Entity manager supposed to work with any entity, automatically find its repository and call its method, whatever
 * entity type are you passing. This version of ReactiveEntityManager works with reactive streams and observables.
 */
var ReactiveEntityManager = (function () {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    function ReactiveEntityManager(connection) {
        this.connection = connection;
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Gets repository of the given entity.
     */
    ReactiveEntityManager.prototype.getRepository = function (entityClass) {
        return this.connection.getRepository(entityClass);
    };
    /**
     * Gets reactive repository of the given entity.
     */
    ReactiveEntityManager.prototype.getReactiveRepository = function (entityClass) {
        return this.connection.getReactiveRepository(entityClass);
    };
    /**
     * Checks if entity has an id.
     */
    ReactiveEntityManager.prototype.hasId = function (entity) {
        return this.getReactiveRepository(entity.constructor).hasId(entity);
    };
    /**
     * Creates a new query builder that can be used to build an sql query.
     */
    ReactiveEntityManager.prototype.createQueryBuilder = function (entityClass, alias) {
        return this.getReactiveRepository(entityClass).createQueryBuilder(alias);
    };
    /**
     * Creates a new entity. If fromRawEntity is given then it creates a new entity and copies all entity properties
     * from this object into a new entity (copies only properties that should be in a new entity).
     */
    ReactiveEntityManager.prototype.create = function (entityClass, fromRawEntity) {
        return this.getReactiveRepository(entityClass).create(fromRawEntity);
    };
    /**
     * Creates a entities from the given array of plain javascript objects.
     */
    ReactiveEntityManager.prototype.createMany = function (entityClass, copyFromObjects) {
        return this.getReactiveRepository(entityClass).createMany(copyFromObjects);
    };
    /**
     * Creates a new entity from the given plan javascript object. If entity already exist in the database, then
     * it loads it (and everything related to it), replaces all values with the new ones from the given object
     * and returns this new entity. This new entity is actually a loaded from the db entity with all properties
     * replaced from the new object.
     */
    ReactiveEntityManager.prototype.initialize = function (entityClass, object) {
        return this.getReactiveRepository(entityClass).initialize(object);
    };
    /**
     * Merges two entities into one new entity.
     */
    ReactiveEntityManager.prototype.merge = function (entity1, entity2) {
        return this.getReactiveRepository(entity1).merge(entity1, entity2);
    };
    /**
     * Persists (saves) a given entity in the database.
     */
    ReactiveEntityManager.prototype.persist = function (entity) {
        // todo: extra casting is used strange tsc error here, check later maybe typescript bug
        return this.getReactiveRepository(entity.constructor).persist(entity);
    };
    /**
     * Removes a given entity from the database.
     */
    ReactiveEntityManager.prototype.remove = function (entity) {
        // todo: extra casting is used strange tsc error here, check later maybe typescript bug
        return this.getReactiveRepository(entity.constructor).remove(entity);
    };
    /**
     * Finds entities that match given conditions.
     */
    ReactiveEntityManager.prototype.find = function (entityClass, conditionsOrFindOptions, options) {
        if (conditionsOrFindOptions && options) {
            return this.getReactiveRepository(entityClass).find(conditionsOrFindOptions, options);
        }
        else if (conditionsOrFindOptions) {
            return this.getReactiveRepository(entityClass).find(conditionsOrFindOptions);
        }
        else {
            return this.getReactiveRepository(entityClass).find();
        }
    };
    /**
     * Finds entities that match given conditions.
     */
    ReactiveEntityManager.prototype.findAndCount = function (entityClass, conditionsOrFindOptions, options) {
        if (conditionsOrFindOptions && options) {
            return this.getReactiveRepository(entityClass).findAndCount(conditionsOrFindOptions, options);
        }
        else if (conditionsOrFindOptions) {
            return this.getReactiveRepository(entityClass).findAndCount(conditionsOrFindOptions);
        }
        else {
            return this.getReactiveRepository(entityClass).findAndCount();
        }
    };
    /**
     * Finds first entity that matches given conditions.
     */
    ReactiveEntityManager.prototype.findOne = function (entityClass, conditionsOrFindOptions, options) {
        if (conditionsOrFindOptions && options) {
            return this.getReactiveRepository(entityClass).findOne(conditionsOrFindOptions, options);
        }
        else if (conditionsOrFindOptions) {
            return this.getReactiveRepository(entityClass).findOne(conditionsOrFindOptions);
        }
        else {
            return this.getReactiveRepository(entityClass).findOne();
        }
    };
    /**
     * Finds entity with given id.
     */
    ReactiveEntityManager.prototype.findOneById = function (entityClass, id, options) {
        return this.getReactiveRepository(entityClass).findOneById(id, options);
    };
    /**
     * Executes raw SQL query and returns raw database results.
     */
    ReactiveEntityManager.prototype.query = function (query) {
        return Rx.Observable.fromPromise(this.connection.driver.query(query));
    };
    /**
     * Wraps given function execution (and all operations made there) in a transaction.
     */
    ReactiveEntityManager.prototype.transaction = function (runInTransaction) {
        var _this = this;
        var runInTransactionResult;
        return Rx.Observable.fromPromise(this.connection.driver
            .beginTransaction()
            .then(function () { return runInTransaction(); })
            .then(function (result) {
            runInTransactionResult = result;
            return _this.connection.driver.endTransaction();
        })
            .then(function () { return runInTransactionResult; }));
    };
    return ReactiveEntityManager;
}());
exports.ReactiveEntityManager = ReactiveEntityManager;
//# sourceMappingURL=ReactiveEntityManager.js.map